#include "cmaudio.h"

#define DRIVER_VERSION "0.69"
#define TO44K   147L / 160L
#define TO48K   160L / 147L
#define OLD_TIMER

const char * card_names[] = {
    "Intel ICH4",
    "VIA VT8233",
    "SiS 7012",
    "SiS 7018",
    "ALI 5451",
    "nVIDIA nForce",
    "Intel ICH2",
    "VIA VT82686",
    "Intel ICH5",
    "AMD 8111",
    "ATI IXP"
};

static const unsigned int chip_id[] = {
	AC97_CM9738_41,
	AC97_CM9739_61,
	AC97_CM9761_78,
	AC97_CM9761_82,
	AC97_CM9761_83,
	AC97_CM9780_69,
};

static const char * chip_names[] = {
    "CMI9738",
    "CMI9739",
    "CMI9761 (078)",
    "CMI9761 (082)",
    "CMI9761 (083)",
    "CMI9780",
};

static const unsigned int MulRatio[101][3] = {
	{17, 17, 17}, {7, 9, 13}, {6, 8, 12}, {6, 7, 8}, {5, 7, 11}, {5, 6, 9},
	{5, 6, 7}, {4, 8, 9}, {4, 6, 10}, {4, 6, 7}, {4, 5, 8}, {4, 5, 6},
	{4, 5, 6}, {3, 8, 10}, {3, 7, 8}, {3, 6, 7}, {3, 5, 9}, {3, 5, 7},
	{3, 5, 6}, {3, 4, 9}, {3, 4, 7}, {3, 4, 6}, {3, 4, 5}, {3, 4, 5},
	{3, 4, 5}, {2, 17, 17}, {2, 7, 9}, {2, 6, 8}, {2, 6, 7}, {2, 5, 7},
	{2, 5, 6}, {2, 5, 6}, {2, 4, 8}, {2, 4, 6}, {2, 4, 6}, {2, 4, 5},
	{2, 4, 5}, {2, 4, 5}, {2, 3, 8}, {2, 3, 7}, {2, 3, 6}, {2, 3, 5},
	{2, 3, 5}, {2, 3, 5}, {2, 3, 4}, {2, 3, 4}, {2, 3, 4}, {2, 3, 4},
	{2, 3, 4}, {2, 3, 4}, {1, 17, 17}, {1, 7, 9}, {1, 6, 8}, {1, 6, 7},
	{1, 5, 7}, {1, 5, 6}, {1, 5, 6}, {1, 4, 8}, {1, 4, 6}, {1, 4, 6}, {1, 4, 5},
	{1, 4, 5}, {1, 4, 5}, {1, 3, 8}, {1, 3, 7}, {1, 3, 6}, {1, 3, 5}, {1, 3, 5},
	{1, 3, 5}, {1, 3, 4}, {1, 3, 4}, {1, 3, 4}, {1, 3, 4}, {1, 3, 4}, {1, 3, 4},
	{1, 2, 17}, {1, 2, 7}, {1, 2, 6}, {1, 2, 6}, {1, 2, 5}, {1, 2, 5}, {1, 2, 5},
	{1, 2, 4}, {1, 2, 4}, {1, 2, 4}, {1, 2, 4}, {1, 2, 4}, {1, 2, 4}, {1, 2, 3},
	{1, 2, 3}, {1, 2, 3}, {1, 2, 3}, {1, 2, 3}, {1, 2, 3}, {1, 2, 3}, {1, 2, 3},
	{1, 2, 3}, {1, 2, 3}, {1, 2, 3}, {1, 2, 3}, {0, 17, 17}
};

static struct cmedia_card *devs = NULL;
#ifdef MUTEUNMUTE
int pcm_can_unmute = 0;
int pcm_volume = 0;
#endif

/*---------------------------------------------------------------------------------------*/
static inline long GetNewSampleLen(INDEX doppler_rate, INDEX index, long OldLen)
{
    long dwTemp = OldLen;
    if (OldLen <= 0)
	   return 0;
    return ((dwTemp << 16) - index.s.fract) / doppler_rate.dw;
}

static short *RecordingSamplingRateConvert
(
    short   *pSource,		// Source Data starting pointer
    short   *MaxSource,		// Source Data maximun pointer
    short   *pTarget,		// Target Data starting pointer
    long     MaxCount,		// Target buffer samples count
    INDEX    doppler_rate,	// Sampling rate ratio
    INDEX   *index,		// Current sampling rate ratio
    unsigned nChannels,		// Channels count
    long     *Count		// Process sample count
)
{
    long normal_doppler;
    unsigned i, j;
    int fract;
    short *p1, *p2, *Source, *Target;

    Source = pSource; Target = pTarget;
    normal_doppler = 0x10000;
    *Count = 0;

    if (doppler_rate.dw == normal_doppler) {
	// 48000, just copy to target
	i = (MaxSource - Source) / 2 / nChannels;
	if (i > MaxCount)
		i = MaxCount;
	memcpy(Target, Source, i * 2 * nChannels);
	*Count = i;
	return (Source + i * nChannels);
    } else {
	p1 = Source + nChannels * index->s.whole;
	p2 = p1 + nChannels;

	while ((p2 + nChannels) < MaxSource && *Count < MaxCount) {
	    // process the sample rate convert
	    fract = index->s.fract;
	    for (j = 0; j < nChannels; j++) {
		*Target = (short)((*p1 * (normal_doppler - fract) + *p2 * fract) >> 16);
		Target++;
		p1++;
		p2++;
	    }

	    (*Count)++;
	    index->dw += doppler_rate.dw;
	    p1 = Source + nChannels * index->s.whole;
	    p2 = p1 + nChannels;
	 }
    }

    index->s.whole = 0; // reset to the first sample
    return p1; // return the last source buffer position
}

/*---------------------------------------------------------------------------------------*/
static inline void ProcessData(struct cmedia_channel *channel, int rec, void *TargetBuf, long count)
{
    struct cmedia_card *card = devs;
    short *ptrSource, *ptrTarget;
    long ProcessCount, cnt;
    long value;

    if (count <= 0)
	   return;

    if (!rec) {
	if (channel->out_count < count)
		count = channel->out_count;
	count >>= 1;

	ptrSource = (short *)(channel->outbuf + channel->out_swptr);
	ptrTarget = (short *)TargetBuf;
	while (count) {
	    cnt = 0;
	    ProcessCount = channel->out_total_bytes - channel->out_swptr;
	    ProcessCount >>= 1;
	    while (count && ProcessCount) {
		if (card->IsAC3 == 1) {
		    if (channel->IsAC3 == 0)
			value = *ptrTarget;
		    else
			value = *ptrSource;
		} else {
		    value = *ptrSource;
		    value += *ptrTarget;
		}

		if (value > 32767)
			*ptrTarget = 32767;
		else if (value < -32768)
			*ptrTarget = -32768;
		else
			*ptrTarget = value;
		ptrSource++;
		ProcessCount--;
		ptrTarget++;
		cnt++;
		count--;
	    }

	    channel->out_swptr = (channel->out_swptr + (cnt << 1)) % channel->out_total_bytes;
	    ptrSource = (short *)(channel->outbuf + channel->out_swptr);
	}
    } else {
	if ((channel->in_total_bytes - channel->in_count) < count)
	    count = channel->in_total_bytes - channel->in_count;
	count >>= 1;

	ptrSource = (short *)TargetBuf;
	ptrTarget = (short *)(channel->inbuf + channel->in_hwptr);

	while (count) {
	    cnt = 0;
	    ProcessCount = channel->in_total_bytes - channel->in_hwptr;
	    ProcessCount >>= 1;
	    while (count && ProcessCount) {
		*ptrTarget = *ptrSource;
		ptrSource++;
		ptrTarget++;
		cnt++;
		count--;
		ProcessCount--;
	    }
	    channel->in_hwptr = (channel->in_hwptr + (cnt << 1)) % channel->in_total_bytes;
	    ptrTarget = (short *)(channel->inbuf + channel->in_hwptr);
	}
    }
}

#ifdef TIMER_ALT
static void update_timer(struct cmedia_card *card, unsigned long timeout)
{
	static void cmedia_timeout(unsigned long ptr);

	del_timer(&card->timer);
	card->timer.function = cmedia_timeout;
	card->timer.data = (unsigned long)card;
	card->timer.expires = jiffies + timeout;
	add_timer(&card->timer);
}
#endif

static void cmedia_timeout(unsigned long ptr)
{
    struct cmedia_card *card;
    struct cmedia_channel *channel;
    int i, left = 0, right = 0, left1 = 0, right1 = 0;
    int l1, l2, l3, r1, r2, r3, l4, l5, l6, r4, r5, r6;
    unsigned long flags;
    short *ptrTarget, *ptrSource, *MaxPos;
    long InProcessCount1 = 0, InProcessCount2 = 0, OutProcessCount1 = 0, OutProcessCount2 = 0;
    long value = 0;

    l1 = r1 = l4 = r4 = 0;
    l2 = l3 = r2 = r3 = l5 = l6 = r5 = r6 = 17;

    card = (struct cmedia_card *) devs;
    spin_lock_irqsave(&card->lock, flags);
    if (!card->dma_state)
	   goto RESET_TIMER;

    card->UpdatePtr(card);

    if (card->dma_state & ADC_RUNNING) {
	if (card->pcmin_swptr < card->pcmin_hwptr) {
	    InProcessCount1 = card->pcmin_hwptr - card->pcmin_swptr;
	    InProcessCount2 = 0;
	} else {
	    InProcessCount1 = PCMIN_DMABUF_SIZE - card->pcmin_swptr;
	    InProcessCount2 = card->pcmin_hwptr;
	}

	if ((card->ac97_vendor_id == AC97_CM9739_61 ||
	     card->ac97_vendor_id == AC97_CM9761_78 ||
	     card->ac97_vendor_id == AC97_CM9761_82 ||
	     card->ac97_vendor_id == AC97_CM9761_83)
	&& card->ReadRegisterWord(card, AC97_RECORD_SELECT) == AC97_RECMUX_STEREO_MIX &&
	    (card->dma_state & DAC_RUNNING) && card->IsAC3 == 0) {
	    ptrTarget = (short *)(card->pcmin_rawbuf + card->pcmin_swptr);
	    MaxPos = (short *)(card->pcmout_rawbuf + PCMOUT_DMABUF_SIZE);
	    ptrSource = (short *)(card->pcmout_rawbuf + card->pcmout_swptr);
	    for (i = 0; i < InProcessCount1 >> 1; i++) {
		value = *ptrSource;
		value += *ptrTarget;
		if (value > 32767)
			*ptrTarget = 32767;
		else if (value < -32768)
			*ptrTarget = -32768;
		else
			*ptrTarget = value;
		ptrTarget++;
		ptrSource++;
		if ((i % 2) == 1)
			ptrSource += 4;
		if (ptrSource >= MaxPos)
			ptrSource = card->pcmout_rawbuf;
	    }

	    ptrTarget = (short *)(card->pcmin_rawbuf);
	    for (i = 0; i < InProcessCount2 >> 1; i++) {
		value = *ptrSource;
		value += *ptrTarget;
		if (value > 32767)
			*ptrTarget = 32767;
		else if (value < -32768)
			*ptrTarget = -32768;
		else
			*ptrTarget = value;
		ptrTarget++;
		ptrSource++;
		if ((i % 2) == 1)
			ptrSource += 4;
		if (ptrSource >= MaxPos)
			ptrSource = card->pcmout_rawbuf;
	    }
	}
    }

    if (card->dma_state & DAC_RUNNING) {
	if (card->pcmout_swptr < card->pcmout_hwptr) {
	    memset((card->pcmout_rawbuf + card->pcmout_swptr), 0, (card->pcmout_hwptr - card->pcmout_swptr));
	    OutProcessCount1 = card->pcmout_hwptr - card->pcmout_swptr;
	    OutProcessCount2 = 0;
	} else {
	    memset((card->pcmout_rawbuf + card->pcmout_swptr), 0, (PCMOUT_DMABUF_SIZE - card->pcmout_swptr));
	    OutProcessCount1 = PCMOUT_DMABUF_SIZE - card->pcmout_swptr;
	    memset((card->pcmout_rawbuf), 0, card->pcmout_hwptr);
	    OutProcessCount2 = card->pcmout_hwptr;
	}
    }

    for (i = 0; i < NR_VR_CH; i++) {
	channel = card->channels[i];
	if (channel && (channel->enable & ADC_RUNNING)) {
	    ptrTarget = (short *)(card->pcmin_rawbuf + card->pcmin_swptr);
	    ProcessData(channel, 1, ptrTarget, InProcessCount1);

	    if (InProcessCount2) {
		ptrTarget = card->pcmin_rawbuf;
		ProcessData(channel, 1, ptrTarget, InProcessCount2);
	    }

	    if ((channel->in_count + InProcessCount1 + InProcessCount2) < channel->in_total_bytes)
		channel->in_count += (InProcessCount1 + InProcessCount2);
	    else
		channel->in_count = channel->in_total_bytes;
	}

	if (channel && (channel->enable & DAC_RUNNING)) {
	    ptrTarget = (short *)(card->pcmout_rawbuf + card->pcmout_swptr);
	    ProcessData(channel, 0, ptrTarget, OutProcessCount1);

	    if (OutProcessCount2) {
		ptrTarget = card->pcmout_rawbuf;
		ProcessData(channel, 0, ptrTarget, OutProcessCount2);
	    }

	    if (channel->out_count > (OutProcessCount1 + OutProcessCount2))
		channel->out_count -= (OutProcessCount1 + OutProcessCount2);
	    else
		channel->out_count = 0;
	}

    }

    if (card->ac97_vendor_id == AC97_CM9738_41) {
	left1 = card->mixer_state[SOUND_MIXER_PCM] & 0x00ff;
	right1 = (card->mixer_state[SOUND_MIXER_PCM] & 0xff00) >> 8;
	left = left1;
	right = right1;
	if (left == 0 && left1 != 0)
		left = 1;
	if (right == 0 && right1 != 0)
		right = 1;
	l1 = MulRatio[left][0];
       	l2 = MulRatio[left][1];
       	l3 = MulRatio[left][2];
	r1 = MulRatio[right][0];
       	r2 = MulRatio[right][1];
       	r3 = MulRatio[right][2];

	l4 = MulRatio[20][0];
       	l5 = MulRatio[20][1];
       	l6 = MulRatio[20][2];
	r4 = MulRatio[20][0];
       	r5 = MulRatio[20][1];
       	r6 = MulRatio[20][2];
    } else if ((card->ac97_vendor_id == AC97_CM9739_61 ||
		card->ac97_vendor_id == AC97_CM9761_78 ||
		card->ac97_vendor_id == AC97_CM9761_82 ||
		card->ac97_vendor_id == AC97_CM9761_83)
		&& (card->IsAC3 == 0)) {
	left1 = card->mixer_state[SOUND_MIXER_VOLUME] & 0x00ff;
	right1 = (card->mixer_state[SOUND_MIXER_VOLUME] & 0xff00) >> 8;
	left = left1;
       	right = right1;
	if (left == 0 && left1 != 0)
		left = 1;
	if (right == 0 && right1 != 0)
		right = 1;
	l1 = MulRatio[left][0];
       	l2 = MulRatio[left][1];
       	l3 = MulRatio[left][2];
	r1 = MulRatio[right][0];
       	r2 = MulRatio[right][1];
       	r3 = MulRatio[right][2];

	left1 = card->mixer_state[SOUND_MIXER_PCM] & 0x00ff;
	right1 = (card->mixer_state[SOUND_MIXER_PCM] & 0xff00) >> 8;
	left = left1;
       	right = right1;
	if (left == 0 && left1 != 0)
		left = 1;
	if (right == 0 && right1 != 0)
		right = 1;
	l4 = MulRatio[left][0];
      	l5 = MulRatio[left][1];
      	l6 = MulRatio[left][2];
	r4 = MulRatio[right][0];
       	r5 = MulRatio[right][1];
       	r6 = MulRatio[right][2];
    }

    if ((card->dma_state & DAC_RUNNING) && (card->IsAC3 == 0)) {
	ptrTarget = (short *)(card->pcmout_rawbuf + card->pcmout_swptr);
	for (i = OutProcessCount1 >> 2; i > 0; i--) {
	    ptrTarget[0] = (ptrTarget[0] >> l1) + (ptrTarget[0] >> l2) + (ptrTarget[0] >> l3);
	    ptrTarget[1] = (ptrTarget[1] >> r1) + (ptrTarget[1] >> r2) + (ptrTarget[1] >> r3);
	    ptrTarget[0] = (ptrTarget[0] >> l4) + (ptrTarget[0] >> l5) + (ptrTarget[0] >> l6);
	    ptrTarget[1] = (ptrTarget[1] >> r4) + (ptrTarget[1] >> r5) + (ptrTarget[1] >> r6);
	    ptrTarget += 2;
	}

	ptrTarget = (short *)(card->pcmout_rawbuf);
	for (i = OutProcessCount2 >> 2; i > 0; i--) {
	    ptrTarget[0] = (ptrTarget[0] >> l1) + (ptrTarget[0] >> l2) + (ptrTarget[0] >> l3);
	    ptrTarget[1] = (ptrTarget[1] >> r1) + (ptrTarget[1] >> r2) + (ptrTarget[1] >> r3);
	    ptrTarget[0] = (ptrTarget[0] >> l4) + (ptrTarget[0] >> l5) + (ptrTarget[0] >> l6);
	    ptrTarget[1] = (ptrTarget[1] >> r4) + (ptrTarget[1] >> r5) + (ptrTarget[1] >> r6);
	    ptrTarget += 2;
	}
    }

    card->ConvertPCMOUT(card);

    if (card->dma_state & DAC_RUNNING)
	    card->pcmout_swptr = card->pcmout_hwptr;
    if (card->dma_state & ADC_RUNNING)
	    card->pcmin_swptr = card->pcmin_hwptr;

    for (i = 0; i < NR_VR_CH; i++) {
	channel = card->channels[i];
	if (channel && (channel->enable & ADC_RUNNING))
		wake_up(&channel->in_wait);
	if (channel && (channel->enable & DAC_RUNNING))
		wake_up(&channel->out_wait);
    }

RESET_TIMER:
    if (devs)
#ifdef TIMER_ALT
	update_timer(card, card->process_time / 30);
#else
	mod_timer(&card->timer, jiffies + card->process_time / 30);
#endif
    spin_unlock_irqrestore(&card->lock, flags);
}

/*---------------------------------------------------------------------------------------*/
/* allocate DMA buffer, playback and recording buffer should be allocated seperately */
static int __init alloc_dmabuf(struct cmedia_card *card)
{
    void *rawbuf = NULL;
    int order;
    struct page *page, *pend;

    //--------------------------------------------------------
    /* alloc enough to satisfy the PCMOUT_DMABUF_SIZE */
    rawbuf = NULL;
    for (order = DMABUF_DEFAULTORDER; order >= DMABUF_MINORDER; order--) {
	if ((PAGE_SIZE << order) > PCMOUT_DMABUF_SIZE)
		continue;
	order++;
	if ((rawbuf = pci_alloc_consistent(card->pci_dev,
					   PAGE_SIZE << order,
					   &card->pcmout_dma_handle)))
		break;
    }

    if (!rawbuf)
	    return -ENOMEM;

    card->pcmout_rawbuf = rawbuf;
    card->pcmout_buforder = order;

    /* now mark the pages as reserved; otherwise remap_page_range doesn't do what we want */
    pend = virt_to_page(rawbuf + (PAGE_SIZE << order) - 1);
    for (page = virt_to_page(rawbuf); page <= pend; page++)
	    SetPageReserved(page);

    //--------------------------------------------------------
    /* alloc enough to satisfy the PCMOUT_DMABUF_SIZE */
    rawbuf = pci_alloc_consistent(card->pci_dev,
					   PAGE_SIZE << order,
					   &card->dmaout_dma_handle);

    if (!rawbuf)
	    return -ENOMEM;

    card->dmaout_rawbuf = rawbuf;

    /* now mark the pages as reserved; otherwise remap_page_range doesn't do what we want */
    pend = virt_to_page(rawbuf + (PAGE_SIZE << order) - 1);
    for (page = virt_to_page(rawbuf); page <= pend; page++)
	    SetPageReserved(page);

    //--------------------------------------------------------
    /* alloc enough to satisfy the PCMIN_DMABUF_SIZE */
    rawbuf = NULL;
    for (order = DMABUF_DEFAULTORDER; order >= DMABUF_MINORDER; order--) {
	if ((PAGE_SIZE << order) > PCMIN_DMABUF_SIZE)
		continue;
	order++;
	if ((rawbuf = pci_alloc_consistent(card->pci_dev,
					   PAGE_SIZE << order,
					   &card->pcmin_dma_handle)))
		break;
    }

    if (!rawbuf)
	    return -ENOMEM;

    card->pcmin_rawbuf = rawbuf;
    card->pcmin_buforder = order;

    /* now mark the pages as reserved; otherwise remap_page_range doesn't do what we want */
    pend = virt_to_page(rawbuf + (PAGE_SIZE << order) - 1);
    for (page = virt_to_page(rawbuf); page <= pend; page++)
	    SetPageReserved(page);

    return 0;
}

/* free DMA buffer */
static void __exit dealloc_dmabuf(struct cmedia_card *card)
{
    struct page *page, *pend;

    if (card->pcmout_rawbuf) {
	/* undo marking the pages as reserved */
	pend = virt_to_page(card->pcmout_rawbuf + (PAGE_SIZE << card->pcmout_buforder) - 1);
	for (page = virt_to_page(card->pcmout_rawbuf); page <= pend; page++)
		ClearPageReserved(page);
	pci_free_consistent(card->pci_dev, PAGE_SIZE << card->pcmout_buforder, card->pcmout_rawbuf, card->pcmout_dma_handle);
    }

    card->pcmout_rawbuf = NULL;

    if (card->dmaout_rawbuf) {
	/* undo marking the pages as reserved */
	pend = virt_to_page(card->dmaout_rawbuf + (PAGE_SIZE << card->pcmout_buforder) - 1);
	for (page = virt_to_page(card->dmaout_rawbuf); page <= pend; page++)
		ClearPageReserved(page);
	pci_free_consistent(card->pci_dev, PAGE_SIZE << card->pcmout_buforder, card->dmaout_rawbuf, card->dmaout_dma_handle);
    }

    card->dmaout_rawbuf = NULL;

    if (card->pcmin_rawbuf) {
	/* undo marking the pages as reserved */
	pend = virt_to_page(card->pcmin_rawbuf + (PAGE_SIZE << card->pcmin_buforder) - 1);
	for (page = virt_to_page(card->pcmin_rawbuf); page <= pend; page++)
		ClearPageReserved(page);
	pci_free_consistent(card->pci_dev, PAGE_SIZE << card->pcmin_buforder, card->pcmin_rawbuf, card->pcmin_dma_handle);
    }

    card->pcmin_rawbuf = NULL;
}

/*---------------------------------------------------------------------------------------*/
/* in this loop, dmabuf.count signifies the amount of data that is waiting to be copied to
   the user's buffer.  it is filled by the dma machine and drained by this loop. */
static ssize_t cmedia_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
    struct cmedia_channel *channel = (struct cmedia_channel *)file->private_data;
    struct cmedia_card *card = channel->card;
    ssize_t ret;
    unsigned long flags;
    long swptr;
    long cnt, new_cnt, left, i;
    int opened, samplesize;
    void *NewP;
    short *DataP;
    unsigned char *NewDataP;
    int RetryCount;

    samplesize = channel->PhysicalChannels << 1;

    if (ppos != &file->f_pos)
	    return -ESPIPE;
    if (!(card->dma_state & ADC_RUNNING))
	    card->StartADC(card);
    if (!access_ok(VERIFY_WRITE, buffer, count))
	    return -EFAULT;
    spin_lock_irqsave(&card->lock, flags);
    if (channel->opened) {
	channel->enable |= ADC_RUNNING;
	channel->trigger |= PCM_ENABLE_INPUT;
    }
    spin_unlock_irqrestore(&card->lock, flags);
    ret = left = 0;
    if (channel->bits == 8)
	    count *= 2;
    RetryCount = 0;

    while (count > 0) {
	spin_lock_irqsave(&card->lock, flags);
	cnt = channel->in_count;
	swptr = channel->in_swptr;
	if (cnt > (channel->in_total_bytes - swptr))
		cnt = channel->in_total_bytes - swptr;
	opened = channel->opened;
	spin_unlock_irqrestore(&card->lock, flags);
	if (!opened)
		return -EFAULT;

	if ((cnt <= 0) && (RetryCount < 10)) {
	    unsigned long tmo;
	    if (!(card->dma_state & ADC_RUNNING)) {
		  channel->trigger |= PCM_ENABLE_INPUT;
		  card->StartADC(card);
	    }

	    RetryCount++;
	    if ((RetryCount >= 10) && (file->f_flags & O_NONBLOCK)) {
		if (!ret)
			ret = -EAGAIN;
		return ret;
	    }

	    tmo = card->process_time / 3;
	    interruptible_sleep_on_timeout(&channel->in_wait, tmo);

	    if (signal_pending(current)) {
		ret = ret ? ret : -ERESTARTSYS;
		return ret;
	    }

	    continue;
	} else
		RetryCount = 0;

	new_cnt = cnt;

	if (channel->rate != 48000) {
	    // Copy to the temp buffer until temp buffer is full
	    if (cnt > channel->pcmin_tempbuf1_count)
		    cnt = new_cnt = channel->pcmin_tempbuf1_count;

	    memcpy(channel->pcmin_tempbuf1 + TEMPBUF_SIZE - channel->pcmin_tempbuf1_count,
		   channel->inbuf + swptr, cnt);
	    channel->pcmin_tempbuf1_count -= cnt;

	    if (channel->pcmin_tempbuf1_count <= 0) {
		NewP = RecordingSamplingRateConvert(
			   channel->pcmin_tempbuf1,
			   channel->pcmin_tempbuf1 + TEMPBUF_SIZE,
			   channel->pcmin_tempbuf,
			   count / samplesize,
			   channel->pcmin_doppler_rate,
			   &(channel->pcmin_index),
			   channel->PhysicalChannels,
			   &cnt);
		cnt *= samplesize;

		memcpy(channel->pcmin_tempbuf1, NewP, channel->pcmin_tempbuf1 + TEMPBUF_SIZE - NewP);
		channel->pcmin_tempbuf1_count = NewP - channel->pcmin_tempbuf1;

		if (channel->bits != 8) {
		    if (copy_to_user(buffer, channel->pcmin_tempbuf, cnt)) {
			if (!ret)
				ret = -EFAULT;
			return ret;
		    }
		} else {
		    DataP = (short *)channel->pcmin_tempbuf;
		    NewDataP = (unsigned char *)channel->pcmin_tempbuf;
		    for (i = 0; i < cnt / 2; i++) {
			*NewDataP = (unsigned char)((*DataP >> 8) ^ 0x80);
			DataP++;
			NewDataP++;
		    }
		    if (copy_to_user(buffer, channel->pcmin_tempbuf, cnt / 2)) {
			if (!ret)
				ret = -EFAULT;
			return ret;
		    }
		}
	    } else
		cnt = 0;

	    swptr = (swptr + new_cnt) % channel->in_total_bytes;
	} else {
	    if (cnt > count)
		    cnt = new_cnt = count;

	    if (channel->bits != 8) {
		if (copy_to_user(buffer, channel->inbuf + swptr, cnt)) {
		    if (!ret)
			    ret = -EFAULT;
		    return ret;
		}
	    } else {
		DataP = (short *)(channel->inbuf + swptr);
		NewDataP = (unsigned char *)(channel->inbuf + swptr);
		for (i = 0; i < cnt / 2; i++) {
		    *NewDataP = (unsigned char)((*DataP >> 8) ^ 0x80);
		    DataP++;
		    NewDataP++;
		}
		if (copy_to_user(buffer, channel->inbuf + swptr, cnt / 2)) {
		    if (!ret)
			    ret = -EFAULT;
		    return ret;
		}
	    }

	    swptr = (swptr + new_cnt) % channel->in_total_bytes;
	}

	spin_lock_irqsave(&card->lock, flags);
	channel->in_swptr = swptr;
	channel->in_count -= new_cnt;
	spin_unlock_irqrestore(&card->lock, flags);

	count -= cnt;
	if (channel->bits != 8)
		buffer += cnt;
       	else
		buffer += cnt / 2;
	if (channel->bits != 8)
		ret += cnt;
       	else
		ret += cnt / 2;
    }
    return ret;
}

/* in this loop, dmabuf.count signifies the amount of data that is waiting to be dma to
   the soundcard.  it is drained by the dma machine and filled by this loop. */
static ssize_t cmedia_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
    struct cmedia_channel *channel = (struct cmedia_channel *)file->private_data;
    struct cmedia_card *card = channel->card;
    ssize_t ret;
    unsigned long flags, i;
    long hwptr = 0;
    long cnt, new_cnt, left;
    int opened, samplesize;
    short *ptrSource, *ptrTarget, *DataP;
    unsigned char *NewDataP;
    void *NewP;
    int RetryCount;

    if (ppos != &file->f_pos)
	    return -ESPIPE;

    if (!(card->dma_state & DAC_RUNNING))
	    card->StartDAC(card);
    if (!access_ok(VERIFY_READ, buffer, count))
	    return -EFAULT;
    if (channel->bits == 8)
	    count *= 2;
    spin_lock_irqsave(&card->lock, flags);
    if (channel->opened) {
	if (!(channel->enable&DAC_RUNNING)) {
	    if (card->FourChannelCount < 0)
		    card->FourChannelCount = 0;
	    if (card->SixChannelCount < 0)
		    card->SixChannelCount = 0;

	    if (channel->IsAC3 == 1) {
		card->IsAC3++;
		if (card->ac97_vendor_id == AC97_CM9739_61 ||
		    card->ac97_vendor_id == AC97_CM9761_78 ||
		    card->ac97_vendor_id == AC97_CM9761_82 ||
		    card->ac97_vendor_id == AC97_CM9761_83) {
			MaskRegOr(card, AC97REG_RESERVED2, 0x0002);
			MaskRegOr(card, AC97_PCMOUT_VOL, AC97_MUTE);
		}
	    }

	    if (channel->PhysicalChannels == 4) {
		card->FourChannelCount++;
#if AUTO_PHONEJACK_SHARE
		if (card->FourChannelCount == 1 && card->SixChannelCount == 0) {
		    card->LineAsRear(card, 1);
		    card->MicAsCBass(card, 0);
	    	}
#endif
	    } else if (channel->PhysicalChannels == 6) {
		card->SixChannelCount++;
#if AUTO_PHONEJACK_SHARE
		if (card->FourChannelCount == 0 && card->SixChannelCount == 1) {
		    card->LineAsRear(card, 1);
		}

		if (card->SixChannelCount == 1) {
		    card->MicAsCBass(card, 1);
		}
#endif
	    }

	    channel->enable |= DAC_RUNNING;
	    channel->trigger |= PCM_ENABLE_OUTPUT;
	}
    }
    spin_unlock_irqrestore(&card->lock, flags);
    ret = 0;
    samplesize = channel->PhysicalChannels << 1;
    RetryCount = 0;

    while (count > 0) {
	spin_lock_irqsave(&card->lock, flags);
	cnt = channel->out_total_bytes - channel->out_count;
	hwptr = channel->out_hwptr;

	// this is to make the copy_to_user simpler below
	if (cnt > channel->out_total_bytes - hwptr)
		cnt = channel->out_total_bytes - hwptr;
	opened = channel->opened;
	spin_unlock_irqrestore(&card->lock, flags);
	if (!opened)
		return -EFAULT;

	if ((cnt == 0) && (RetryCount < 10)) {
	    unsigned long tmo;

	    RetryCount++;
	    if ((RetryCount >= 10) && (file->f_flags & O_NONBLOCK)) {
		if (!ret)
			ret = -EAGAIN;
		return ret;
	    }

	    tmo = card->process_time / 3;
	    interruptible_sleep_on_timeout(&channel->out_wait, tmo);

	    if (signal_pending(current)) {
		if (!ret)
			ret = -ERESTARTSYS;
		return ret;
	    }
	    continue;
	} else
		RetryCount = 0;

	cnt = cnt * channel->PhysicalChannels / 6;
	new_cnt = cnt;

	if (channel->rate != 48000) {
	    left = TEMPBUF_SIZE - channel->pcmout_leftsize;
	    if (left > count)
		    cnt = count;
	    else
		    cnt = left;

	    if (channel->bits != 8) {
		if (copy_from_user(channel->pcmout_tempbuf + channel->pcmout_leftsize, buffer, cnt)) {
		    if (!ret)
			    ret = -EFAULT;
		    return ret;
		}
	    } else {
		if (copy_from_user(channel->pcmout_tempbuf + channel->pcmout_leftsize, buffer, cnt / 2)) {
		    if (!ret)
			    ret = -EFAULT;
		    return ret;
		}
		DataP = (short *)(channel->pcmout_tempbuf + channel->pcmout_leftsize + cnt - 2);
		NewDataP = (unsigned char *)(channel->pcmout_tempbuf + channel->pcmout_leftsize + cnt / 2 - 1);
		for (i = 0; i < cnt / 2; i++) {
		    *DataP = (short)(((short)*NewDataP << 8) - 32767);
		    DataP--;
		    NewDataP--;
		}
	    }

	    NewP = RecordingSamplingRateConvert(
			channel->pcmout_tempbuf,
			channel->pcmout_tempbuf + cnt + channel->pcmout_leftsize,
			channel->outbuf + hwptr,
			new_cnt / samplesize,
			channel->pcmout_doppler_rate,
			&(channel->pcmout_index),
			channel->PhysicalChannels,
			&new_cnt);
	    new_cnt *= samplesize;

	    channel->pcmout_leftsize = channel->pcmout_tempbuf + cnt + channel->pcmout_leftsize - NewP;
	    if (left > 0)
		    memcpy(channel->pcmout_tempbuf, NewP, channel->pcmout_leftsize);
	    else
		    channel->pcmout_leftsize = 0;
	} else {
	    if (cnt > count)
		    cnt = new_cnt = count;

	    if (channel->bits != 8) {
		if (copy_from_user(channel->outbuf + hwptr, buffer, cnt)) {
		    if (!ret)
			    ret = -EFAULT;
		    return ret;
		}
	    } else {
		if (copy_from_user(channel->outbuf + hwptr, buffer, cnt / 2)) {
		    if (!ret)
			    ret = -EFAULT;
		    return ret;
		}
		DataP = (short *)(channel->outbuf + hwptr + cnt - 2);
		NewDataP = (unsigned char *)(channel->outbuf + hwptr + cnt / 2 - 1);
		for (i = 0; i < cnt / 2; i++) {
		    *DataP = (short)(((short)*NewDataP << 8) - 32767);
		    DataP--;
		    NewDataP--;
		}
	    }
	}

	// Process to the correct position
	{
	    switch (channel->PhysicalChannels) {
	    case 1:
		ptrSource = (unsigned short *)(channel->outbuf + hwptr + new_cnt - 2);
		ptrTarget = (unsigned short *)(channel->outbuf + hwptr + new_cnt * 6 - 2);

		for (i = new_cnt / 2; i > 0; i--) {
		    ptrTarget[0] =
		    ptrTarget[-1] =
		    ptrTarget[-2] =
		    ptrTarget[-3] = 0;
		    ptrTarget[-4] =
		    ptrTarget[-5] = ptrSource[0];
		    ptrTarget -= 6;
		    ptrSource--;
		}
		break;
	    case 2:
		ptrSource = (unsigned short *)(channel->outbuf + hwptr + new_cnt - 2);
		ptrTarget = (unsigned short *)(channel->outbuf + hwptr + new_cnt * 3 - 2);

		for (i = new_cnt / 4; i > 0; i--) {
		    ptrTarget[0] =
		    ptrTarget[-1] =
		    ptrTarget[-2] =
		    ptrTarget[-3] = 0;
		    ptrTarget[-4] = ptrSource[0];
		    ptrTarget[-5] = ptrSource[-1];
		    ptrTarget -= 6;
		    ptrSource -= 2;
		}
		break;
	    case 4:
		ptrSource = (unsigned short *)(channel->outbuf + hwptr + new_cnt - 2);
		ptrTarget = (unsigned short *)(channel->outbuf + hwptr + new_cnt * 3 / 2 - 2);

		for (i = new_cnt / 8; i > 0; i--) {
		    ptrTarget[0] =
		    ptrTarget[-1] = 0;
		    ptrTarget[-2] = ptrSource[0];
		    ptrTarget[-3] = ptrSource[-1];
		    ptrTarget[-4] = ptrSource[-2];
		    ptrTarget[-5] = ptrSource[-3];
		    ptrTarget -= 6;
		    ptrSource -= 4;
		}
		break;
	    case 6:
		break;
	    }
	}

	hwptr = (hwptr + new_cnt * 6 / channel->PhysicalChannels) % channel->out_total_bytes;

	spin_lock_irqsave(&card->lock, flags);
	channel->out_hwptr = hwptr;
	channel->out_count += new_cnt * 6 / channel->PhysicalChannels;
	spin_unlock_irqrestore(&card->lock, flags);

	count -= cnt;
	if (channel->bits != 8)
		buffer += cnt;
	else
		buffer += cnt / 2;
	if (channel->bits != 8)
		ret += cnt;
	else
		ret += cnt / 2;
	//printk(KERN_DEBUG "cmedia_write , count = %d ret = %d cnt = %ld\n", count, ret, cnt);
    }

    return ret;
}

static int cmedia_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    struct cmedia_channel *channel = (struct cmedia_channel *)file->private_data;
    struct cmedia_card *card = channel->card;
    unsigned long flags;
    audio_buf_info abinfo;
    long swptr = 0;
    long cnt;
    //count_info cinfo;
    int val, ret;
    //int ret;
    INDEX index;
#if 0
    const char *ioctl_cmd[] = {
	    "RESET",
	    "SYNC",
	    "SPEED",
	    "STEREO",
	    "GETBLKSIZE",
	    "SETFMT",
	    "CHANNELS",
	    "WRITE_FILTER",
	    "POST",
	    "SUBDIVIDE",
	    "SETFRAGMENT",
	    "GETFMTS",
	    "GETOSPACE",
	    "GETISPACE",
	    "NONBLOCK",
	    "GETCAPS",
	    "TRIGGER",
	    "GETIPTR",
	    "GETOPTR",
	    "MAPINBUF",
	    "MAPOUTBUF",
	    "SETSYNCRO",
	    "SETDUPLEX",
	    "GETODELAY"
    };

    printk(KERN_DEBUG "CMedia Audio: cmedia_ioctl, arg = 0x%x, cmd = %d ", arg ? *(int *)arg : 0, cmd & 0xff);
    if ((cmd & 0xff) < 24)
	printk("%s\n", ioctl_cmd[cmd & 0xff]);
    else
    	printk("\n");
#endif

    switch (cmd) {
    case OSS_GETVERSION:
	return put_user(SOUND_VERSION, (int *)arg);

    case SNDCTL_DSP_RESET:
#if 0
	if (file->f_mode & FMODE_WRITE) {
		card->DrainDAC(card);
		synchronize_irq();
	}
	if (file->f_mode & FMODE_READ) {
		card->StopADC(card);
		synchronize_irq();
	}
#endif
	return 0;

    case SNDCTL_DSP_SYNC:
#if 0
	if (file->f_mode & FMODE_WRITE)
		return card->DrainDAC(card);
#endif
	return 0;

    case SNDCTL_DSP_SPEED: /* set smaple rate */
	if (get_user(val, (int *)arg))
		return -EFAULT;
	if (val >= 0) {
		if (val > 48000)
			val = 48000;
		else if (val < 8000)
			val = 8000;
		channel->rate = val;
		channel->pcmout_doppler_rate.dw = val;
		channel->pcmout_doppler_rate.dw <<= 16;
		channel->pcmout_doppler_rate.dw /= 48000;
		channel->pcmout_index.dw = 0;
		channel->pcmin_doppler_rate.dw = 48000;
		channel->pcmin_doppler_rate.dw <<= 16;
		channel->pcmin_doppler_rate.dw /= val;
		channel->pcmin_index.dw = 0;
	}
	return put_user(channel->rate, (int *)arg);

    case SNDCTL_DSP_STEREO: /* set stereo or mono channel */
	if (get_user(val, (int *)arg))
		return -EFAULT;
	spin_lock_irqsave(&card->lock, flags);
	if (val) {
		channel->fmt |= CMEDIA_FMT_STEREO;
		channel->PhysicalChannels = 2;
       	} else {
		channel->fmt &= ~CMEDIA_FMT_STEREO;
		channel->PhysicalChannels = 1;
       	}
	spin_unlock_irqrestore(&card->lock, flags);

	return 0;

    case SNDCTL_DSP_GETBLKSIZE:
	return put_user(PAGE_SIZE, (int *)arg);

    case SNDCTL_DSP_GETFMTS: /* Returns a mask of supported sample format*/
	return put_user(AFMT_S16_LE | AFMT_AC3, (int *)arg);

    case SNDCTL_DSP_SETFMT: /* Select sample format */
	if (get_user(val, (int *)arg))
		return -EFAULT;
	if (!(val == 8 || val == 16 || val == AFMT_AC3))
		return -EINVAL;

	if (val == AFMT_AC3) {
	    spin_lock_irqsave(&card->lock, flags);
	    val = 48000;
	    channel->rate = val;
	    channel->pcmout_doppler_rate.dw = val;
	    channel->pcmout_doppler_rate.dw <<= 16;
	    channel->pcmout_doppler_rate.dw /= 48000;
	    channel->pcmout_index.dw = 0;
	    channel->pcmin_doppler_rate.dw = 48000;
	    channel->pcmin_doppler_rate.dw <<= 16;
	    channel->pcmin_doppler_rate.dw /= val;
	    channel->pcmin_index.dw = 0;
	    channel->PhysicalChannels = 2;
	    channel->IsAC3 = 1;
	    channel->fmt |= CMEDIA_FMT_STEREO;
	    spin_unlock_irqrestore(&card->lock, flags);
	    return put_user(AFMT_AC3, (int *)arg);
	} else {
	    spin_lock_irqsave(&card->lock, flags);
	    channel->IsAC3 = 0;
	    channel->bits = val;
	    spin_unlock_irqrestore(&card->lock, flags);
	    return put_user(val, (int *)arg);
	}

    case SNDCTL_DSP_CHANNELS:
	if (get_user(val, (int *)arg))
		return -EFAULT;
	if (val != 0) {
		if (!(val == 1 || val == 2 || val == 4 || val == 6))
			return -EINVAL;
		if (val > card->output_channels)
			return -EFAULT;

		spin_lock_irqsave(&card->lock, flags);
		channel->PhysicalChannels = val;
		spin_unlock_irqrestore(&card->lock, flags);

		return put_user(val, (int *)arg);
	}
	return put_user(channel->PhysicalChannels, (int *)arg);

    case SNDCTL_DSP_POST: /* the user has sent all data and is notifying us */
	return 0;

    case SNDCTL_DSP_SUBDIVIDE:
	return -EFAULT;

    case SNDCTL_DSP_SETFRAGMENT:
	if (get_user(val, (int *)arg))
		return -EFAULT;
	abinfo.fragsize = val & 0xffff;
	abinfo.fragments = (val >> 16) & 0xffff;
	return 0;

    case SNDCTL_DSP_GETOSPACE:
	if (!(file->f_mode & FMODE_WRITE))
		return -EINVAL;

	cnt = 0;
	abinfo.fragsize = PAGE_SIZE;
	abinfo.fragstotal = (PAGE_SIZE << card->pcmout_buforder) / PAGE_SIZE;
	abinfo.fragments = 0;
	index.dw = 0;

	while (cnt == 0 || abinfo.fragments == 0) {
	    spin_lock_irqsave(&card->lock, flags);
	    cnt = channel->out_total_bytes - channel->out_count;
	    cnt /= 12;
	    cnt = GetNewSampleLen(channel->pcmin_doppler_rate, index, cnt);
	    spin_unlock_irqrestore(&card->lock, flags);
	    cnt *= 2 * channel->PhysicalChannels;
	    abinfo.bytes = cnt;
	    abinfo.fragments = abinfo.bytes / PAGE_SIZE;

	    if (cnt == 0 || abinfo.fragments == 0) {
		unsigned long tmo;
		tmo = card->process_time / 3;
		interruptible_sleep_on_timeout(&channel->out_wait, tmo);
	    }

	    if (signal_pending(current))
		    return -ERESTARTSYS;
	}
	abinfo.bytes = cnt;
	abinfo.fragments = abinfo.bytes / PAGE_SIZE;
	return copy_to_user((void *)arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;

    case SNDCTL_DSP_GETOPTR:
	return -EFAULT;

    case SNDCTL_DSP_GETISPACE:
	return -EINVAL;

	cnt = 0;
	while (cnt == 0) {
	    spin_lock_irqsave(&card->lock, flags);
	    cnt = channel->in_count;
	    swptr = channel->in_swptr;
	    if (cnt > (channel->in_total_bytes - swptr))
		    cnt = channel->in_total_bytes - swptr;
	    spin_unlock_irqrestore(&card->lock, flags);

	    if (cnt == 0) {
		unsigned long tmo;
		tmo = card->process_time / 3;
		interruptible_sleep_on_timeout(&channel->in_wait, tmo);
	    }

	    if (signal_pending(current))
		    return -ERESTARTSYS;
	}
	abinfo.fragsize = PAGE_SIZE;
	abinfo.fragstotal = (PAGE_SIZE << card->pcmin_buforder) / PAGE_SIZE;
	abinfo.bytes = cnt;
	abinfo.fragments = abinfo.bytes / PAGE_SIZE;

	return copy_to_user((void *)arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;

    case SNDCTL_DSP_GETIPTR:
	return -EFAULT;

    case SNDCTL_DSP_NONBLOCK:
	file->f_flags |= O_NONBLOCK;
	return 0;

    case SNDCTL_DSP_GETCAPS:
	return put_user(DSP_CAP_DUPLEX | DSP_CAP_BATCH | DSP_CAP_TRIGGER | DSP_CAP_MULTI, (int *)arg);

    case SNDCTL_DSP_GETTRIGGER:
	val = 0;
	if (file->f_mode & FMODE_READ && channel->enable & ADC_RUNNING)
		val |= PCM_ENABLE_INPUT;
	if (file->f_mode & FMODE_WRITE && channel->enable & DAC_RUNNING)
		val |= PCM_ENABLE_OUTPUT;
	return put_user(val, (int *)arg);

    case SNDCTL_DSP_SETTRIGGER:
	if (get_user(val, (int *)arg))
		return -EFAULT;
	if (file->f_mode & FMODE_READ) {
		if (val & PCM_ENABLE_INPUT)
			card->StartADC(card);
		else
			card->StopADC(card);
	}
	if (file->f_mode & FMODE_WRITE) {
		if (val & PCM_ENABLE_OUTPUT)
			card->StartDAC(card);
		else
			card->DrainDAC(card);
	}
	return 0;

    case SNDCTL_DSP_SETDUPLEX:
	return 0;

    case SNDCTL_DSP_GETODELAY:
	if (!(file->f_mode & FMODE_WRITE))
		return -EINVAL;
	spin_lock_irqsave(&card->lock, flags);
	card->UpdatePtr(card);
	val = channel->out_count;
	spin_unlock_irqrestore(&card->lock, flags);
	return put_user(val, (int *)arg);

    case SOUND_PCM_READ_RATE:
	return put_user(channel->rate, (int *)arg);

    case SOUND_PCM_READ_CHANNELS:
	return put_user(2, (int *)arg);

    case SOUND_PCM_READ_BITS:
	return put_user(16, (int *)arg);

    case SNDCTL_DSP_MAPINBUF:
    case SNDCTL_DSP_MAPOUTBUF:
    case SNDCTL_DSP_SETSYNCRO:
    case SOUND_PCM_WRITE_FILTER:
    case SOUND_PCM_READ_FILTER:
	return -EINVAL;

    case SNDCTL_DSP_GETCHANNELMASK:
	return put_user(DSP_BIND_FRONT | DSP_BIND_SURR | DSP_BIND_CENTER_LFE | DSP_BIND_SPDIF, (int *)arg);

	/* --- extened ioctls --- */
    case SNDCTL_DSP_LINEINASREAR:
	if (get_user(val, (int *)arg))
		return -EFAULT;
	if (card->output_channels < 4)
		return -EINVAL;
#if FORCE_LINEINASREAR_MODE
	return -EINVAL;
#endif
	ret = card->LineAsRear(card, val);
	return ret >= 0 ? put_user(ret, (int *)arg) : ret;

    case SNDCTL_DSP_MICASCENTERBASS:
	if (get_user(val, (int *)arg))
		return -EFAULT;
	if (card->output_channels < 6)
		return -EINVAL;
#if FORCE_MICASBASS_MODE
	return -EINVAL;
#endif
	ret = card->MicAsCBass(card, val);
	return ret >= 0 ? put_user(ret, (int *)arg) : ret;

    case SNDCTL_DSP_XEAR:
	if (get_user(val, (int *)arg))
		return -EFAULT;
	if (card->output_channels < 4)
		return -EINVAL;
#if FORCE_XEAR_MODE
	return -EINVAL;
#endif
	ret = card->Xear(card, val);
	return ret >= 0 ? put_user(ret, (int *)arg) : ret;

    case SNDCTL_DSP_MICSELECT2:
	if (get_user(val, (int *)arg))
		return -EFAULT;
#if FORCE_MICSELECT2_MODE
	return -EINVAL;
#endif
	ret = card->MicSel2(card, val);
	return ret >= 0 ? put_user(ret, (int *)arg) : ret;

    case SNDCTL_DSP_MIX2SURROUND:
	if (get_user(val, (int *)arg))
		return -EFAULT;

	ret = card->Mix2Surr(card, val);
	return ret >= 0 ? put_user(ret, (int *)arg) : ret;

    case SNDCTL_DSP_MIC_BOOST:
	if (get_user(val, (int *)arg))
		return -EFAULT;

	ret = card->MicBoostSel(card, val);
	return ret >= 0 ? put_user(ret, (int *)arg) : ret;

    case SNDCTL_DSP_FMIC:
	if (get_user(val, (int *)arg))
		return -EFAULT;

	ret = card->FMicSel(card, val);
	return ret >= 0 ? put_user(ret, (int *)arg) : ret;
      }

    return -EINVAL;
}

static int cmedia_init_timer(struct cmedia_card *card)
{
    init_timer(&card->timer);
    card->timer.function = cmedia_timeout;
    card->timer.data = (unsigned long)card;
    add_timer(&card->timer);

    return 0;
}

static int cmedia_open(struct inode *inode, struct file *file)
{
	int i = 0;
	struct cmedia_card *card = devs;
	struct cmedia_channel *channel = NULL;
	unsigned long flags;

	/* find an avaiable virtual channel (instance of /dev/dsp) */
	if (card != NULL) {
	    for (i = 0; i < 50 && card->initializing; i++) {
		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(card->process_time / 20);
	    }

	    for (i = 0; i < NR_VR_CH && !card->initializing; i++) {
		if (card->channels[i] == NULL) {
		    channel = (struct cmedia_channel *) kmalloc(sizeof(struct cmedia_channel), GFP_KERNEL);
		    if (channel == NULL)
			   return -ENOMEM;
		    memset(channel, 0, sizeof(struct cmedia_channel));
		    goto found_virt;
		}
	    }
	}

	/* no more virtual channel avaiable */
	if (!channel)
		return -ENODEV;

found_virt:

	/* initialize the virtual channel */
	channel->virt = i;
	channel->PhysicalChannels = 2;
	channel->card = card;
	init_waitqueue_head(&channel->in_wait);
	init_waitqueue_head(&channel->out_wait);
	init_MUTEX(&channel->open_sem);
	file->private_data = channel;
	channel->rate = 48000;
	channel->pcmout_doppler_rate.dw = 65536;
	channel->pcmout_index.dw = 0;
	channel->pcmout_leftsize = 0;
	channel->pcmin_doppler_rate.dw = 65536;
	channel->pcmin_index.dw = 0;
	channel->fmt = CMEDIA_FMT_16BIT | CMEDIA_FMT_STEREO;
	channel->IsAC3 = 0;
	channel->open_mode |= file->f_mode & (FMODE_READ | FMODE_WRITE);

	/* allocate hardware channels */
	if ((channel->outbuf = (void *) kmalloc(PCMOUT_DMABUF_SIZE, GFP_KERNEL)) == NULL)
		goto no_outbuf;

	channel->out_hwptr = channel->out_swptr = channel->out_count = 0;
	channel->out_total_bytes = PCMOUT_DMABUF_SIZE;

	if ((channel->inbuf = (void *) kmalloc(PCMIN_DMABUF_SIZE, GFP_KERNEL)) == NULL)
		goto no_inbuf;

	channel->in_hwptr = channel->in_swptr = channel->in_count = 0;
	channel->in_total_bytes = PCMIN_DMABUF_SIZE;

	// for sampling rate covert
	if ((channel->pcmout_tempbuf = (void *) kmalloc(TEMPBUF_SIZE, GFP_KERNEL)) == NULL)
		goto no_pcmout_tempbuf;

	if ((channel->pcmin_tempbuf = (void *) kmalloc(TEMPBUF_SIZE, GFP_KERNEL)) == NULL)
		goto no_pcmin_tempbuf;

	if ((channel->pcmin_tempbuf1 = (void *) kmalloc(TEMPBUF_SIZE, GFP_KERNEL)) == NULL)
		goto no_pcmin_tempbuf1;

	channel->pcmin_tempbuf1_count = TEMPBUF_SIZE;

	channel->opened = 1;

#ifndef OLD_TIMER
	if (card->pcmout_count == 0 && card->pcmin_count == 0)
		cmedia_init_timer(card);
#endif
	spin_lock_irqsave(&card->lock, flags);
	if (file->f_mode & FMODE_WRITE) {
		card->pcmout_count++;
		channel->trigger |= PCM_ENABLE_OUTPUT;
	}
	if (file->f_mode & FMODE_READ) {
		card->pcmin_count++;
		channel->trigger |= PCM_ENABLE_INPUT;
	}
	card->channels[i] = channel;
	spin_unlock_irqrestore(&card->lock, flags);
	return 0;

no_pcmin_tempbuf1:
	kfree (channel->pcmin_tempbuf);
no_pcmin_tempbuf:
	kfree (channel->pcmout_tempbuf);
no_pcmout_tempbuf:
	kfree (channel->inbuf);
no_inbuf:
	kfree (channel->outbuf);
no_outbuf:
	kfree (channel);
	return -ENOMEM;
}

static int cmedia_release(struct inode *inode, struct file *file)
{
    struct cmedia_channel *channel = (struct cmedia_channel *)file->private_data;
    struct cmedia_card *card = channel->card;
    unsigned long flags;
    unsigned long tmo;
    int enable;

    if (channel->PhysicalChannels == 4) {
	card->FourChannelCount--;
#if AUTO_PHONEJACK_SHARE
	if (card->FourChannelCount == 0 && card->SixChannelCount == 0)
	    card->LineAsRear(card, 0);
#endif
    } else if (channel->PhysicalChannels == 6) {
	card->SixChannelCount--;
#if AUTO_PHONEJACK_SHARE
	if (card->FourChannelCount == 0 && card->SixChannelCount == 0)
	    card->LineAsRear(card, 0);

	if (card->SixChannelCount == 0)
	    card->MicAsCBass(card, 0);
#endif
    }

    spin_lock_irqsave(&card->lock, flags);
    card->channels[channel->virt] = NULL;
    enable = channel->enable;
    channel->enable = channel->opened = 0;
    if (file->f_mode & FMODE_WRITE)
	   card->pcmout_count--;
    if (file->f_mode & FMODE_READ)
	   card->pcmin_count--;

    /* stop DMA state machine and free DMA buffers/channels */
    if ((enable & ADC_RUNNING) && (card->pcmin_count == 0))
	   card->StopADC(card);
    if ((enable & DAC_RUNNING) && (card->pcmout_count == 0))
	   card->DrainDAC(card);
    card->IsAC3 -= channel->IsAC3;
    if (card->IsAC3 <= 0) {
	if (card->ac97_vendor_id == AC97_CM9761_78 ||
	    card->ac97_vendor_id == AC97_CM9761_82 ||
	    card->ac97_vendor_id == AC97_CM9761_83) {
	    MaskRegAnd(card, AC97REG_RESERVED2, ~0x0002);

	    if (card->mixer_state[SOUND_MIXER_PCM] != 0)
		MaskRegAnd(card, AC97_PCMOUT_VOL, ~AC97_MUTE);
	}
	card->IsAC3 = 0;
    }
    spin_unlock_irqrestore(&card->lock, flags);

    tmo = card->process_time / 3;
    if (file->f_mode & FMODE_WRITE)
	   interruptible_sleep_on_timeout(&channel->out_wait, tmo);
    else
	   interruptible_sleep_on_timeout(&channel->in_wait, tmo);
#ifndef OLD_TIMER
    if (card->pcmout_count == 0 && card->pcmin_count == 0)
	del_timer_sync(&card->timer);
#endif

    kfree(channel->outbuf);
    kfree(channel->inbuf);
    kfree(channel->pcmout_tempbuf);
    kfree(channel->pcmin_tempbuf);
    kfree(channel->pcmin_tempbuf1);
    kfree(channel);

    return 0;
}

static /*const*/ struct file_operations cmedia_audio_fops = {
	owner:		THIS_MODULE,
	llseek:		no_llseek,
	read:		cmedia_read,
	write:		cmedia_write,
	ioctl:		cmedia_ioctl,
	open:		cmedia_open,
	release:	cmedia_release,
};

/* OSS /dev/mixer file operation methods */

static int cmedia_open_mixdev(struct inode *inode, struct file *file)
{
    int i;
    int minor = MINOR(inode->i_rdev);
    struct cmedia_card *card = devs;

    if (card != NULL) {
	for (i = 0; i < 50 && card->initializing; i++) {
	    current->state = TASK_UNINTERRUPTIBLE;
	    schedule_timeout(card->process_time / 20);
	}

	if (!card->initializing) {
	    if (card->dev_mixer == minor) {
		file->private_data = card;
		return 0;
	    }
	}
    }

    return -ENODEV;
}

static int cmedia_ioctl_mixdev(struct inode *inode, struct file *file, unsigned int cmd,
				unsigned long arg)
{
    int i, val = 0;
    struct cmedia_card *card = devs;

    if (cmd == SOUND_MIXER_INFO) {
	mixer_info info;
	for (i = 0; i < sizeof(chip_id) / sizeof(chip_id[0]); i++) {
		if (card->ac97_vendor_id == chip_id[i]) {
			strncpy(info.id, chip_names[i], sizeof(info.id));
			strncpy(info.name, chip_names[i], sizeof(info.name));
			break;
		}
	}
	info.modify_counter = card->modify_counter;
	if (copy_to_user((void *)arg, &info, sizeof(info)))
		return -EFAULT;
	return 0;
    }

    if (cmd == SOUND_OLD_MIXER_INFO) {
	_old_mixer_info info;
	for (i = 0; i < sizeof(chip_id) / sizeof(chip_id[0]); i++) {
		if (card->ac97_vendor_id == chip_id[i]) {
			strncpy(info.id, chip_names[i], sizeof(info.id));
			strncpy(info.name, chip_names[i], sizeof(info.name));
			break;
		}
	}
	if (copy_to_user((void *)arg, &info, sizeof(info)))
		return -EFAULT;
	return 0;
    }

    if (_IOC_TYPE(cmd) != 'M' || _SIOC_SIZE(cmd) != sizeof(int))
	    return -EINVAL;

    if (cmd == OSS_GETVERSION)
	    return put_user(SOUND_VERSION, (int *)arg);

    if (_SIOC_DIR(cmd) == _SIOC_READ) {
	switch (_IOC_NR(cmd)) {
	case SOUND_MIXER_RECSRC: /* give them the current record source */
	    if (!card->RecMask_io)
		    val = 0;
	    else
		    val = card->RecMask_io(card, 1, 0);
	    break;

	case SOUND_MIXER_DEVMASK: /* give them the supported mixers */
	    val = card->supported_mixers;
	    break;

	case SOUND_MIXER_RECMASK: /* Arg contains a bit for each supported recording source */
	    val = card->record_sources;
	    break;

	case SOUND_MIXER_STEREODEVS: /* Mixer channels supporting stereo */
	    val = card->stereo_mixers;
	    break;

	case SOUND_MIXER_CAPS:
	    val = SOUND_CAP_EXCL_INPUT;
	    break;

	default: /* read a specific mixer */
	    i = _IOC_NR(cmd);

	    if (!card_supported_mixer(card, i))
		    return -EINVAL;

	    /* do we ever want to touch the hardware? */
	    /* val = card->ReadRegisterWord(card, i); */
	    val = card->mixer_state[i];
	    break;
	}

	return put_user(val, (int *)arg);
    }

    if (_SIOC_DIR(cmd) == (_SIOC_WRITE | _SIOC_READ)) {
	card->modify_counter++;
	if (get_user(val, (int *)arg))
		return -EFAULT;

	switch (_IOC_NR(cmd)) {
	case SOUND_MIXER_RECSRC: /* Arg contains a bit for each recording source */
	    if (!card->RecMask_io)
		    return -EINVAL;
	    if (!val)
		    return 0;
	    if (!(val &= card->record_sources))
		    return -EINVAL;

	    card->RecMask_io(card, 0, val);

	    return 0;

	default: /* write a specific mixer */
	    i = _IOC_NR(cmd);

	    if (!card_supported_mixer(card, i))
		    return -EINVAL;

	    card->Set_Mixer(card, i, val);

	    return 0;
	}
    }

    return -EINVAL;
}

static /*const*/ struct file_operations cmedia_mixer_fops = {
	owner:		THIS_MODULE,
	llseek:		no_llseek,
	ioctl:		cmedia_ioctl_mixdev,
	open:		cmedia_open_mixdev,
};

/* CMedia Audio initialisation. */
static int __init cmedia_audio_init(struct cmedia_card *card)
{
    u16 val;
    u16 wTemp;
    int i;

    if (!card->Initial || card->Initial(card) < 1) {
	printk(KERN_ERR "CMedia Audio: couldn't initial!\n");
	return 0;
    }

    switch (card->ac97_vendor_id) {
	    case AC97_CM9738_41:
		card->InitialMixer = InitialMixerCMI9738;
		card->output_channels = 4;
		break;
	    case AC97_CM9739_61:
		card->InitialMixer = InitialMixerCMI9739;
		card->output_channels = 6;
		break;
	    case AC97_CM9761_78:
		card->InitialMixer = InitialMixerCMI9761;
		card->output_channels = 6;
		break;
	    case AC97_CM9761_82:
		card->InitialMixer = InitialMixerCMI9761;
		card->output_channels = 6;
		/* find A or B */
		val = card->ReadRegisterWord(card, AC97REG_RESERVED);
		/* set page 1 */
		wTemp = (val & ~0x0f) | 1;
		card->WriteRegisterWord(card, AC97REG_RESERVED, wTemp);
		wTemp = card->ReadRegisterWord(card, 0x60);
		if (wTemp & 1)
			card->revision = 'B';
		else
			card->revision = 'A';
		/* restore page */
		card->WriteRegisterWord(card, AC97REG_RESERVED, val);
		break;
	    case AC97_CM9761_83:
		card->InitialMixer = InitialMixerCMI9761;
		card->output_channels = 6;
		break;
	    case AC97_CM9780_69:
		card->InitialMixer = InitialMixerCMI9761;
		card->output_channels = 6;
		break;
	    default:
		card->InitialMixer = NULL;
		card->output_channels = 2;
		break;
    }

    if (!card->InitialMixer) {
	printk(KERN_ERR "%x not supported\n", card->ac97_vendor_id);
	return 0;
    }

    card->InitialMixer(card);

    card->LineInAsRear = 0;
    card->MicAsCenterBass = 0;
#if FORCE_XEAR_MODE
    card->XearMode = 1;
#else
    card->XearMode = 0;
#endif
    card->FourChannelCount = 0;
    card->SixChannelCount = 0;

    card->SetSpeakers(card, 6);
#if FORCE_LINEINASREAR_MODE
    card->LineAsRear(card, 1);
#endif

#if FORCE_MICSELECT2_MODE
    card->MicSelect2 = 1;
    MaskRegOr(card, AC97_GENERAL_PURPOSE, AC97_GP_MS);
#else
    card->MicSelect2 = 0;
    MaskRegAnd(card, AC97_GENERAL_PURPOSE, ~AC97_GP_MS);
#endif

#if FORCE_MICASBASS_MODE
    card->MicAsCBass(card, 1);
#endif

    if ((card->dev_mixer = register_sound_mixer(&cmedia_mixer_fops, -1)) < 0) {
	printk(KERN_ERR "CMedia Audio: couldn't register mixer!\n");
	return 0;
    }

    // Set GPIO and PC BEEP
    if (card->ac97_vendor_id == AC97_CM9739_61 ||
	card->ac97_vendor_id == AC97_CM9761_78 ||
       	card->ac97_vendor_id == AC97_CM9761_82 ||
       	card->ac97_vendor_id == AC97_CM9761_83) {
	MaskRegAnd(card, AC97REG_MULTI_CHANNEL_CTRL, ~0x8000);

	card->WriteRegisterWord(card, AC97REG_GPIO_SETUP, 0x0100);
	card->WriteRegisterWord(card, AC97REG_GPIO_STATUS, 0x0020);

	val = card->ReadRegisterWord(card, AC97REG_GPIO_SETUP);
	val = card->ReadRegisterWord(card, AC97REG_GPIO_STATUS);
    }

    // display chip name
    for (i = 0; i < sizeof(chip_id) / sizeof(chip_id[0]); i++) {
	if (card->ac97_vendor_id == chip_id[i]) {
		printk(KERN_INFO "%s", chip_names[i]);
		break;
	}
    }
    if (card->ac97_vendor_id == AC97_CM9761_82)
	    printk("%c\n", card->revision);
    else
	    printk("\n");
    return 1;
}

/* install the driver, we do not allocate hardware channel nor DMA buffer now, they are defered
   until "ACCESS" time (in prog_dmabuf called by open/read/write/ioctl/mmap) */

static int __devinit cmedia_probe(struct pci_dev *pci_dev, const struct pci_device_id *pci_id)
{
    struct cmedia_card *card;
    unsigned long flags;

    printk(KERN_INFO "CMedia Audio, version " DRIVER_VERSION ", " __TIME__ " " __DATE__ "\n");

    if (pci_enable_device(pci_dev))
	    return -EIO;

    if (pci_set_dma_mask(pci_dev, CMEDIA_DMA_MASK)) {
	printk(KERN_ERR "CMedia Audio: architecture does not support 32bit PCI busmaster DMA\n");
	return -ENODEV;
    }

    if ((card = kmalloc(sizeof(struct cmedia_card), GFP_KERNEL)) == NULL) {
	printk(KERN_ERR "CMedia Audio: out of memory\n");
	return -ENOMEM;
    }
    memset(card, 0, sizeof(*card));

    card->initializing = 1;
    card->iobase = pci_resource_start(pci_dev, 1);
    card->opbase = pci_resource_start(pci_dev, 0);
    card->pci_dev = pci_dev;
    card->pci_id = pci_id->device;
    card->driver_data = pci_id->driver_data;
    card->process_time = HZ;
    spin_lock_init(&card->lock);
    devs = card;

    pci_set_master(pci_dev);

    printk(KERN_INFO "CMedia Audio: %s found at IO 0x%04lx and 0x%04lx\n", card_names[pci_id->driver_data], card->iobase, card->opbase);

    if (alloc_dmabuf(card)) {
	kfree(card);
	return -ENOMEM;
    }

    /* point to correct functions */
    switch (pci_id->driver_data) {
    case INTELICH2:
    case INTELICH4:
    case INTELICH5:
    case INTELICH6:
    case SIS7012:
    case NVIDIA_NFORCE:
    case AMD8111:
	card->RequestRegion = RequestRegion_INTEL_ICHx;
	card->Initial = Initial_INTEL_ICHx;
	break;

    case ALI5451:
	card->RequestRegion = RequestRegion_ALI_5451;
	card->Initial = Initial_ALI_5451;
	break;

    case VIA8233:
    case VIA82686:
	card->RequestRegion = RequestRegion_VIA_82cxxx;
	card->Initial = Initial_VIA_82cxxx;
	break;

    case SIS7018:
	card->RequestRegion = RequestRegion_SiS_7018;
	card->Initial = Initial_SiS_7018;
	break;

    case ATIIXP:
	card->RequestRegion = RequestRegion_ATI_IXP;
	card->Initial = Initial_ATI_IXP;
	break;

    default:
	printk("PCI %x not supported\n", pci_id->driver_data);
	card->RequestRegion = NULL;
	card->Initial = NULL;
	break;
    }

    /* claim our iospace and irq */
    if (!card->RequestRegion || !card->RequestRegion(card))
	    goto request_region_fail;

    /* register /dev/dsp */
    if ((card->dev_audio = register_sound_dsp(&cmedia_audio_fops, -1)) < 0)
	    goto register_sound_dsp_fail;

    /* initialize cmedia audio and register /dev/mixer */
    if (cmedia_audio_init(card) <= 0)
	    goto cmedia_audio_init_fail;

    pci_set_drvdata(pci_dev, card);

    /* initail the DMA buffer */
    card->ProgramDMABuffer(card);

    card->initializing = 0;
#ifdef MUTEUNMUTE
    if (card->ac97_vendor_id == AC97_CM9739_61 ||
	card->ac97_vendor_id == AC97_CM9761_78 ||
	card->ac97_vendor_id == AC97_CM9761_82 ||
	card->ac97_vendor_id == AC97_CM9761_83) {
	memset(card->pcmout_rawbuf, card->ac97_vendor_id == AC97_CM9761_78 ? -1755 : 64, PCMOUT_DMABUF_SIZE);
	card->StartDAC(card);
	card->DrainDAC(card);
	/* unmute PCM */
	pcm_can_unmute = 1;
	card->Set_Mixer(card, SOUND_MIXER_PCM, pcm_volume);
    }
#endif
#ifdef OLD_TIMER
    return cmedia_init_timer(card);
#else
    return 0;
#endif

cmedia_audio_init_fail:
	unregister_sound_dsp(card->dev_audio);
register_sound_dsp_fail:
	card->ReleaseRegion(card);
request_region_fail:
	kfree(card);
	return -ENODEV;
}

static void __devinit cmedia_remove(struct pci_dev *pci_dev)
{
    u16 val;
    unsigned long flags;
    struct cmedia_card *card = pci_get_drvdata(pci_dev);

    spin_lock_irqsave(&card->lock, flags);

#ifdef OLD_TIMER
    del_timer_sync(&card->timer);
#endif

    /* stop everything */
    if (card->dma_state & DAC_RUNNING)
	    card->DrainDAC(card);
    if (card->dma_state & ADC_RUNNING)
	    card->StopADC(card);

    devs = NULL;
    spin_unlock_irqrestore(&card->lock, flags);

    // Set GPIO and PC BEEP
    if (card->ac97_vendor_id == AC97_CM9739_61 ||
	card->ac97_vendor_id == AC97_CM9761_78 ||
	card->ac97_vendor_id == AC97_CM9761_82 ||
	card->ac97_vendor_id == AC97_CM9761_83) {
	MaskRegOr(card, AC97REG_MULTI_CHANNEL_CTRL, 0x8000);

	card->WriteRegisterWord(card, AC97REG_GPIO_SETUP, 0x0100);
	card->WriteRegisterWord(card, AC97REG_GPIO_STATUS, 0x0021);

	val = card->ReadRegisterWord(card, AC97REG_GPIO_SETUP);
	val = card->ReadRegisterWord(card, AC97REG_GPIO_STATUS);
    }

    card->ReleaseRegion(card);

    dealloc_dmabuf(card);

    /* unregister audio devices */
    unregister_sound_mixer(card->dev_mixer);

    unregister_sound_dsp(card->dev_audio);

    kfree(card);
    pci_set_drvdata(pci_dev, NULL);
}

MODULE_AUTHOR("Min-Chih Feng");
MODULE_DESCRIPTION("C-Media audio support");
MODULE_LICENSE("GPL");

#define CMEDIA_MODULE_NAME "cmedia_audio"

static struct pci_device_id cmedia_pci_tbl [] __devinitdata = {
    { PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH4, PCI_ANY_ID, PCI_ANY_ID, 0, 0, INTELICH4 },
    { PCI_VENDOR_ID_VIA, PCI_DEVICE_ID_VIA_8233_5,   PCI_ANY_ID, PCI_ANY_ID, 0, 0, VIA8233 },
    { PCI_VENDOR_ID_SI, PCI_DEVICE_ID_SIS_7012,     PCI_ANY_ID, PCI_ANY_ID, 0, 0, SIS7012 },
    { PCI_VENDOR_ID_SI, PCI_DEVICE_ID_SIS_7018,     PCI_ANY_ID, PCI_ANY_ID, 0, 0, SIS7018 },
    { PCI_VENDOR_ID_AL, PCI_DEVICE_ID_AL_5451,     PCI_ANY_ID, PCI_ANY_ID, 0, 0, ALI5451 },
    { PCI_VENDOR_ID_AL, PCI_DEVICE_ID_AL_1563,     PCI_ANY_ID, PCI_ANY_ID, 0, 0, ALI5451 },
    { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_0,     PCI_ANY_ID, PCI_ANY_ID, 0, 0, NVIDIA_NFORCE },
    { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_1,     PCI_ANY_ID, PCI_ANY_ID, 0, 0, NVIDIA_NFORCE },
    { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_2,     PCI_ANY_ID, PCI_ANY_ID, 0, 0, NVIDIA_NFORCE },
    { PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH2, PCI_ANY_ID, PCI_ANY_ID, 0, 0, INTELICH2 },
    { PCI_VENDOR_ID_VIA, PCI_DEVICE_ID_VIA_82C686_5,   PCI_ANY_ID, PCI_ANY_ID, 0, 0, VIA82686 },
    { PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH5, PCI_ANY_ID, PCI_ANY_ID, 0, 0, INTELICH5 },
    { PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6, PCI_ANY_ID, PCI_ANY_ID, 0, 0, INTELICH6 },
    { PCI_VENDOR_ID_AMD, PCI_DEVICE_ID_AMD_8111, PCI_ANY_ID, PCI_ANY_ID, 0, 0, AMD8111 },
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_ATI_IXP, PCI_ANY_ID, PCI_ANY_ID, 0, 0, ATIIXP },
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_ATI_IXP300, PCI_ANY_ID, PCI_ANY_ID, 0, 0, ATIIXP },
    { 0, }
};

MODULE_DEVICE_TABLE (pci, cmedia_pci_tbl);

static struct pci_driver cmaudio_driver = {
    name:	CMEDIA_MODULE_NAME,
    id_table:	cmedia_pci_tbl,
    probe:	cmedia_probe,
    remove:	cmedia_remove
};

static int __init init_cmaudio(void)
{
	return pci_module_init(&cmaudio_driver);
}

static void __exit cmedia_cleanup_module (void)
{
    pci_unregister_driver(&cmaudio_driver);
    printk(KERN_INFO "CMedia Audio exit\n");
}

module_init(init_cmaudio);
module_exit(cmedia_cleanup_module);
